/*
 * QEMU vGlass User Interface (UI)
 *
 * (c) 2015 Assured Information Security, Inc
 * Author: Kyle J. Temkin <temkink@ainfosec.com>
 * Author: Brendan Kerrigan <kerriganb@ainfosec.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//Uncomment the following line to turn on debug messages from
//the vGlass Display Driver Helper.
#define DISPLAY_HELPER_DEBUG

#include "qemu/osdep.h"
#include "qemu-common.h"
#include "qemu/timer.h"
#include "ui/console.h"
#include "ui/input.h"
#include "sysemu/sysemu.h"
#include "qemu/thread.h"
#include <libivc.h>
#include <pv_display_helper.h>
#include <math.h>

#include "ui/vglass.h"

static DisplayChangeListener *dcl;
extern uint32_t xen_domid;

// Try to keep this aligned, currently 32byte aligned
typedef struct xt_input_event
{
    uint32_t type;      // 4
    uint32_t flags;     // 4

    uint32_t absX;      // 4
    uint32_t absY;      // 4

    union
    {
        uint32_t absXMax;   // 4
        float relX;         // 4
    };
    union
    {
        uint32_t absYMax;   // 4
        float relY;         // 4
    };
    uint32_t relZ;      // (Pretty much padding)
    uint32_t keyCode;   // 4
} __attribute__((packed, aligned(32))) xt_input_event;

typedef struct openxt_dcl_context {
    // Current resolution
    unsigned int xres;
    unsigned int yres;
    unsigned int bpp;

    // PV display interface
    struct pv_display_provider *display_provider;
    struct pv_display *display;

    // Internal image object wrapping the display.
    pixman_image_t * pixmap;

    unsigned int display_count;
    unsigned int advertised_cap_count;
    unsigned int advertised_disp_count;

    // True iff we're currently taking priority over a PV driver.
    // See openxt_take_display_if_necessary() below.
    bool take_priority_over_pv;

    // Input channel
    struct libivc_client *input;
    QemuMutex mutex;
    DisplaySurface *surface;
    QEMUTimer *control_reconnect_timer;
    QEMUTimer *display_reconnect_timer;
    QEMUTimer *input_reconnect_timer;
    QEMUPutLEDEntry *led_handler;

    struct dh_add_display *request;
} openxt_dcl_context_t;

openxt_dcl_context_t *odc = NULL;

static void input_reconnect_handler(void *unused, struct libivc_client *input);
static int init_display_provider(DisplayChangeListener *dcl);
static void add_display_request(struct pv_display_provider * provider,
                                struct dh_add_display * request);

/**
 * Internally, we create an pixman image that wraps our IVC buffer, allowing us
 * to take advantage of pixman's image translation capabilities. This function
 * creates the relevant image iff the image does not exist.
 *
 * (We use this function instead of allocating on connection as the allocation
 * /can/ fail, and this this method allows us an easy way to recver.
 *
 * (Note that the pixman image itself does not allocate a backing store for
 *  the image, but simply wraps the existing binary.)
 *
 */
static void create_pixmap_if_necessary(void)
{
    if(odc->pixmap)
        return;

    if(!odc->display)
        return;

    odc->pixmap = pixman_image_create_bits(PIXMAN_x8r8g8b8, OPENXT_MAX_WIDTH, OPENXT_MAX_HEIGHT,
        (uint32_t *)odc->display->framebuffer, OPENXT_MAX_STRIDE);
}

/* Input handling */

static void handle_led_event_change(void *opaque, int led_state)
{
    openxt_dcl_context_t *context = opaque;
    uint32_t ledstate = led_state;
    if(context && context->input) {
        libivc_send(context->input, (char *)&ledstate, sizeof(ledstate));
    }
}

static void vglass_inject_input(xt_input_event *event)
{
    InputButton input_button = 0;
    /* This handling could probably be moved into the QEmuStubdomSink */
    if (event->type != XT_TYPE_KEY) {
        int button = -1;

        switch (event->keyCode) {
            case BTN_LEFT:
                button = INPUT_BUTTON_LEFT;
                break;
            case BTN_RIGHT:
                button = INPUT_BUTTON_RIGHT;
                break;
            case BTN_MIDDLE:
                button = INPUT_BUTTON_MIDDLE;
                break;
        }

        if(button != -1)
            qemu_input_queue_btn(dcl->con, button, event->flags);
    }

    switch (event->type) {
        case XT_TYPE_KEY:
        {
            QKeyCode key_code = qemu_input_to_qcode(event->keyCode);
            qemu_input_event_send_key_qcode(dcl->con, key_code, event->flags);
            break;
        }
        case XT_TYPE_RELATIVE:
        {
            if (event->relZ) {
                if (event->relZ == 1) {
                    input_button = INPUT_BUTTON_WHEEL_DOWN;
                } else if (event->relZ == -1) {
                    input_button = INPUT_BUTTON_WHEEL_UP;
                }
             }

            if (input_button) {
                qemu_input_queue_btn(dcl->con, input_button, true);
                qemu_input_event_sync();
                qemu_input_queue_btn(dcl->con, input_button, false);
                qemu_input_event_sync();
            }

            // qemu_input_queue_rel takes an int, so we must convert from float
            // to int.  vglass uses lround during hotspot tracking, so match
            // that to keep the two in sync.
            qemu_input_queue_rel(dcl->con, INPUT_AXIS_X, lround(event->relX));
            qemu_input_queue_rel(dcl->con, INPUT_AXIS_Y, lround(event->relY));
            break;
        }
        case XT_TYPE_ABSOLUTE:
        {
            if (event->relZ) {
                if (event->relZ == 1) {
                    input_button = INPUT_BUTTON_WHEEL_DOWN;
                } else if (event->relZ == -1) {
                    input_button = INPUT_BUTTON_WHEEL_UP;
                }
             }

            if (input_button) {
                qemu_input_queue_btn(dcl->con, input_button, true);
                qemu_input_event_sync();
                qemu_input_queue_btn(dcl->con, input_button, false);
                qemu_input_event_sync();
            }
            qemu_input_queue_abs(dcl->con, INPUT_AXIS_X, event->absX, 0, event->absXMax);
            qemu_input_queue_abs(dcl->con, INPUT_AXIS_Y, event->absY, 0, event->absYMax);
            break;
        }
        default:
            return;
    }

    qemu_input_event_sync();
}

/* Reconnection support */
static void retry_display_connection(void *opaque)
{
    if(!odc) return;

    if (timer_pending(odc->display_reconnect_timer)) {
        return;
    }

    if(!odc->display_provider) return;
    if(!odc->request) return;

    add_display_request(odc->display_provider, odc->request);

    g_free(odc->request);
    odc->request = NULL;

    return;
}

static void retry_input_connection(void *opaque)
{
    int err = 0;

    // If the VM is currently running (and not asleep), try to reconnect.
    if(runstate_is_running()) {
        err = libivc_connect_with_id(&odc->input, OPENXT_INPUT_BACKEND_DOMID, OPENXT_INPUT_PORT, OPENXT_INPUT_BUFFER_SIZE>>12, xen_domid);

        // If we were able to forge a connection, we're done!
        if(!err) {
            libivc_register_event_callbacks(odc->input, NULL, input_reconnect_handler, NULL);
            return;
        }
    }

    // Otherwise, schedule a retry.
    timer_mod(odc->input_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
}


static void input_reconnect_handler(void *unused, struct libivc_client *input)
{
    (void) unused;

    if (timer_pending(odc->input_reconnect_timer)) {
        return;
    }

    odc->input = NULL;
    libivc_disconnect(input);

    timer_mod(odc->input_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
}

static void reconnect_control_channel(void *opaque)
{
    (void)opaque;

    if (!odc) goto retry_timer;

    if (init_display_provider(dcl)) {
        goto retry_timer;
    }

    return;

retry_timer:
    timer_mod(odc->control_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
}

static void display_provider_error(struct pv_display_provider *provider)
{
    if (timer_pending(odc->control_reconnect_timer)) {
        return;
    }

    // Trigger a reconnect in init_display_provider
    provider->destroy(provider);
    odc->display_provider = NULL;

    timer_mod(odc->control_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
}

/* DisplayChangeListener callbacks */

static void vglass_update(DisplayChangeListener *dcl, int x, int y, int w, int h)
{
    DisplaySurface * surface = qemu_console_surface(dcl->con);

    if (!odc) return;
    if (!odc->display) return;
    if (!odc->display->framebuffer) return;

    create_pixmap_if_necessary();
    if(!odc->pixmap)
    {
        fprintf(stderr, "Failed to create pixmap! Not copying.\n");
        return;
    }

    // Source in the original image. In the best case, this performs
    // a pixman-optimized memcopy; in the worst, it converts color format
    // using SSE2 instructions.
    //
    // All in all, this should always be a well-optimized operation.
    pixman_image_composite(PIXMAN_OP_SRC, surface->image, NULL, odc->pixmap, x, y, 0, 0, x, y, w, h);

    // Mark the relevant region as requiring a redraw.
    odc->display->invalidate_region(odc->display, x, y, w, h);
}

static void full_screen_update(DisplayChangeListener *dcl)
{
    DisplaySurface * surface = qemu_console_surface(dcl->con);
    vglass_update(dcl, 0, 0, surface_width(surface), surface_height(surface));
}

static void vglass_switch(DisplayChangeListener *dcl, DisplaySurface *new_surface)
{
    if (!odc) return;
    if (!odc->display) return;

    odc->bpp  = surface_bytes_per_pixel(new_surface);
    odc->xres = surface_width(new_surface);
    odc->yres = surface_height(new_surface);
    odc->display->change_resolution(odc->display, odc->xres, odc->yres, OPENXT_MAX_STRIDE);
    full_screen_update(dcl);
}

static void vglass_refresh(DisplayChangeListener *dcl)
{
    size_t inputBytesAvailable = 0;
    size_t inputBytesRead = 0;
    xt_input_event event;
    if (!odc) return;
    if (!odc->display_provider) return;

    if (odc->advertised_cap_count < 1) {
        odc->display_provider->advertise_capabilities(odc->display_provider, 1);
        odc->advertised_cap_count++;
    }

    if (odc->input && libivc_isOpen(odc->input)) {
        libivc_getAvailableData(odc->input, &inputBytesAvailable);

        while (inputBytesAvailable >= sizeof(xt_input_event)) {
            memset(&event, 0x00, sizeof(xt_input_event));
            libivc_read(odc->input, (char*)&event, (size_t)sizeof(xt_input_event), &inputBytesRead);
            if (inputBytesRead == sizeof(xt_input_event)) {
                vglass_inject_input(&event);
            }
            inputBytesAvailable -= inputBytesRead;
        }
    }

    graphic_hw_update(dcl->con);
}

static void destroy_context(void)
{
    timer_free(odc->input_reconnect_timer);
    timer_free(odc->control_reconnect_timer);

    if (odc && odc->input) {
        libivc_disconnect(odc->input);
        odc->input = NULL;
    }

    if (odc && odc->pixmap)
        pixman_image_unref(odc->pixmap);

    if (odc && odc->display)
        odc->display->destroy(odc->display);

    if (odc && odc->display_provider)
        odc->display_provider->destroy(odc->display_provider);

    qemu_mutex_destroy(&odc->mutex);

    if (odc) {
        g_free(odc);
        odc = NULL;
    }
}

static void vglass_cleanup(void)
{
    destroy_context();

    if (dcl) {
        g_free(dcl);
    }
}

static void handle_display_disconnected(struct pv_display * display)
{
    if (!odc) return;
    odc->display_count = 0;
}

/* PV Display callbacks */

static void add_display_request(struct pv_display_provider * provider,
                                struct dh_add_display * request)
{
    int err = 0;


    if (!odc) return;
    if (odc->display == NULL) {

        /* Initial display creation */
        err = provider->create_display(provider, &odc->display, request,
              OPENXT_MAX_WIDTH, OPENXT_MAX_HEIGHT,
              OPENXT_MAX_STRIDE, NULL);

        if(err || odc->display == NULL) {
            fprintf(stderr, "Failed to create display(%d)\n", err);
            // Copy the information necessary for a retry
            odc->request = g_malloc0(sizeof(*request));
            memcpy(odc->request, request, sizeof(*request));
            timer_mod(odc->display_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
            return;
        }

        // Create a pixman image wrapping our display object.
        create_pixmap_if_necessary();

        odc->display->set_driver_data(odc->display, odc);
        odc->display->register_fatal_error_handler(odc->display, handle_display_disconnected);
        odc->display_count = 1;
    } else {
        /* Reconnect to an existing display */
        err = odc->display->reconnect(odc->display, request, OPENXT_DISPLAY_DOMAIN);

        odc->display->key = request->key;
        odc->display_count = 1;
    }

    if (odc->display_count && dcl->con)
    {
        DisplaySurface * surface = qemu_console_surface(dcl->con);
        odc->display->change_resolution(odc->display, surface_width(surface), surface_height(surface), OPENXT_MAX_STRIDE);
        full_screen_update(dcl);
    } else {
        odc->display->change_resolution(odc->display, 720, 480, 2880);
    }

    if (odc->input == NULL) {
        pv_display_debug("%d\n", __LINE__);
        err = libivc_connect_with_id(&odc->input, OPENXT_INPUT_BACKEND_DOMID, OPENXT_INPUT_PORT, OPENXT_INPUT_BUFFER_SIZE>>12, xen_domid);
        if (err) {
            fprintf(stderr, "Failed to setup IVC channel to the input server\n");
            timer_mod(odc->input_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + 1000);
        } else {
            libivc_register_event_callbacks(odc->input, NULL, input_reconnect_handler, NULL);
        }
    }
}

static void host_display_changed(struct pv_display_provider *provider,
                                 struct dh_display_info *displays,
                                 uint32_t num_displays)
{
    /*
        Qemu can only deal with one display, and it's size is
        mostly static
    */
    if (!odc) return;
    if (!odc->display_provider) return;
    if (!odc->display_provider->advertise_displays) return;

    displays[0].x = 0;
    displays[0].y = 0;
    displays[0].width = OPENXT_MAX_WIDTH;
    displays[0].height = OPENXT_MAX_HEIGHT;

    odc->display_provider->advertise_displays(provider, displays, 1);
}

static int init_display_provider(DisplayChangeListener *dcl)
{
    int err = 0;

    // QEmu will call openxt_init for both the text console DisplayState,
    // and the graphic console display state. We only want to setup our
    // context once, and use
    if (!odc) return -EINVAL;

    odc->input_reconnect_timer   = timer_new_ms(QEMU_CLOCK_REALTIME, retry_input_connection, odc);
    odc->control_reconnect_timer = timer_new_ms(QEMU_CLOCK_REALTIME, reconnect_control_channel, odc);
    odc->display_reconnect_timer = timer_new_ms(QEMU_CLOCK_REALTIME, retry_display_connection, odc);

#ifdef CONFIG_STUBDOM
    err = create_pv_display_provider(&odc->display_provider, OPENXT_DISPLAY_DOMAIN, OPENXT_CONTROL_PORT);
#else
    err = create_pv_display_provider_with_conn_id(&odc->display_provider, OPENXT_DISPLAY_DOMAIN, OPENXT_CONTROL_PORT, xen_domid);
#endif

    if (err < 0 || odc->display_provider == NULL) {
        fprintf(stderr, "Failed to create PV display provider(%d)\n", err);
        timer_mod(odc->control_reconnect_timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + DH_RETRY_PERIOD);
        return err;
    }

    // Initialize the display counts
    odc->advertised_cap_count = 0;
    odc->display_count = 0;

    // Register a handler for the events (and requests) sent by the Display Handler.
    odc->display_provider->register_host_display_change_handler(odc->display_provider,
        host_display_changed);
    odc->display_provider->register_add_display_request_handler(odc->display_provider,
        add_display_request);
    odc->display_provider->register_fatal_error_handler(odc->display_provider,
        display_provider_error);

    return 0;
}

static DisplayChangeListenerOps vglass_dcl_ops = {
    .dpy_name          = "dhqemu",
    .dpy_gfx_update    = vglass_update,
    .dpy_gfx_switch    = vglass_switch,
    .dpy_refresh       = vglass_refresh,
    //.dpy_mouse_set     = vglass_mouse_warp,
    //.dpy_cursor_define = vglass_mouse_define,
};


/**
 * Initialize the vGlass user interface, which allows the user to display a
 * QEMU guest using the vGlass client viritualization frontend.
 *
 * @param ds The DisplayState to wrap with an vGlass object.
 */
void vglass_display_init(DisplayState * ds, DisplayOptions *opts)
{
    fprintf(stderr, "Using vGlass display...\n");

    dcl = g_malloc0(sizeof(DisplayChangeListener));

    if (!dcl) {
        fprintf(stderr, "Failed to allocate DisplayChangeListener\n");
        exit(-ENOMEM);
    }

    dcl->ops = &vglass_dcl_ops;

    odc = g_malloc0(sizeof(*odc));
    if (!odc) {
        fprintf(stderr, "Failed to allocate OpenXT Display Context\n");
        exit(-ENOMEM);
    }

    //XXX Is this right?
    dcl->con = qemu_console_lookup_by_index(0);

    odc->led_handler = qemu_add_led_event_handler(handle_led_event_change, odc);
    qemu_mutex_init(&odc->mutex);
    register_displaychangelistener(dcl);

    init_display_provider(dcl);

    atexit(vglass_cleanup);
}

static QemuDisplay qemu_display_vglass = {
    .type = DISPLAY_TYPE_DHQEMU,
    .init = vglass_display_init,
};

static void register_vglass(void)
{
    qemu_display_register(&qemu_display_vglass);
}

type_init(register_vglass);
